"""
Class of which object is used to plot data
"""

from PySide2 import QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as Canvas)
from matplotlib import pyplot as pl
from matplotlib.patches import ConnectionPatch, Rectangle
from matplotlib.ticker import AutoMinorLocator
import numpy as np

from sohstationviewer.controller.plottingData import (
    getTitle, getGaps, getTimeTicks, getUnitBitweight, getMassposValueColors)
from sohstationviewer.controller.util import (
    displayTrackingInfo, getVal, apply_convert_factor)

from sohstationviewer.conf import constants
from sohstationviewer.conf.colorSettings import Clr, set_color_mode
from sohstationviewer.conf.dbSettings import dbConf

from sohstationviewer.database import extractData

from sohstationviewer.model.handling_data import trim_downsample_SOHChan


class PlottingWidget(QtWidgets.QScrollArea):

    def __init__(self, parent, tracking_box, name=''):
        """
        Widget to plot SOH channels.
        This inherit QScrollArea allow it to be scroll to see the entire
            drawing area.
        TimePowerSquaredWidget and WaveformWidget also inherit from this widget

        :param parent: QWidget/QMainWindow - widget that contains this plotting
            widget
        :param tracking_box: QTextBrowser - widget to display tracking info
        :param name: str - name of the plotting widget to keep track of what
            sohview is working on
        """
        super().__init__()
        self.name = name
        self.parent = parent
        self.tracking_box = tracking_box
        self.peer_plotting_widgets = [self]
        """
        plot_total: int - total of plots display on this widget
        """
        self.plot_total = 0
        """
        main_widget: QWidget - widget in the scroll area to keep the drawing
            canvas
        """
        self.main_widget = QtWidgets.QWidget(parent)
        """
        axes: [matplotlib.axes.Axes, ] - list of axes for plotting channels
        """
        self.axes = []
        """
        width_base: float - basic width of plotting which can be enlarged to
            the maximum of 1 which is 4x of width_base
        """
        self.width_base = 0.25
        """
        width_base_px: float - width base in pixel which fit the plotting area
        """
        self.width_base_px = 1546.

        """
        ratio_w: float - the ratio for plotting_w. This value changed when the
            width of the scroll area changed or user choose a different
            scaling ratio
        """
        self.ratio_w = 1.
        """
        plotting_w: float - plotting area's width which is set with ratio_w of
            width_base
        """
        self.plotting_w = self.width_base * self.ratio_w

        """
        plotting_bot: float - bottom of a current plot, decrease by plot_h
            return from self.get_height() whenever a new plot is added
        """
        self.plotting_bot = constants.BOTTOM

        """
        plotting_bot_pixel: float - bottom of current plot in pixel
        """
        self.plotting_bot_pixel = constants.BOTTOM_PX
        """
        plotting_l_base: float - The basic left of plotting area
        """
        self.plotting_l_base = 0.04
        """
        plotting_l: float - plotting area's left which is set with ratio_w of
            plotting_l_base.
        TODO: check to see if should keep this left
            unchanged or changed with width of the scrolling area
        """
        self.plotting_l = self.plotting_l_base * self.ratio_w
        """
        font_size: float - font size on plot. With some require bigger font,
            +2 to the font_size
        """
        self.font_size = 7

        """
        fig: matplotlib.pyplot.Figure - figure associate with canvas to add
            axes for plotting
        Set fig's size 50in width, 100in height.
        This is the maximum size of plotting container.
        add_axes will draw proportion to this size.
        The actual view based on size of self.main_widget.
        """
        self.fig = pl.Figure(facecolor='white', figsize=(50, 100))
        self.fig.canvas.mpl_connect('button_press_event',
                                    self.on_button_press_event)

        """
        canvas: matplotlib.backends.backend_qt5agg.FigureCanvasQTAgg - the
            canvas inside main_widget associate with fig
        """
        self.canvas = Canvas(self.fig)
        self.canvas.setParent(self.main_widget)
        self.setWidget(self.main_widget)

        # ================== pre-define ===================
        self.min_x = 0  # left limit of a channel plot
        self.max_x = 0  # right limit of a channel plot

        """
        ruler: ConnectionPatch - a line go from top to bottom to help
            user see the corresponding positions of different channels.
            It shows up at the mouse x position when  Ctrl/Cmd + left click,
            and disappears when zoom_marker1 shows up or when ESC is pressed.
        """
        self.ruler = None
        """
        zoom_marker1: ConnectionPatch - a line go from top to bottom to help
            mark where the zooming area start. It shows up at the mouse x
            position when Shift + left click, and disappears when plot has
            zoomed in or when ESC is pressed.
        """
        self.zoom_marker1 = None
        """
        zoom_marker2: ConnectionPatch - a line go from top to bottom to help
            mark where the zooming area end. It shows up at the mouse x
            position when  Shift + left click after zoom_marker1 shown up,
            and disappears when plot has zoomed in or when ESC is pressed.
        """
        self.zoom_marker2 = None
        """
        zoom_marker1_shown: bool - flag for sohview to know if zoom_marker1 is
            shown/hidden
        """
        self.zoom_marker1_shown = False
        """
        follower: int - connection id to help keeping zoom_marker2 following
            mouse move
        """
        self.follower = None

        """
        plotting_data1: dict that includes 'times', 'data', 'ax'-
            first set of data for plotting. It can be either
            DataTypeModel.__init__.soh_data[key] or
            DataTypeModel.__init__.waveform_data[key]['read_data']
        """
        self.plotting_data1 = {}
        """
        plotting_data2: dict that includes 'times' and 'data'-
            second set of data for plotting. It is
            DataTypeModel.__init__.mass_pos_data[key]
        """
        self.plotting_data2 = {}
        """
        processing_log: [(str,str] - list of processing info and type
        """
        self.processing_log = []
        """
        date_mode: str - format of display date
        """
        self.date_mode = 'YYYY-MM-DD'
        """
        time_ticks_total: int - maximum of total major ticks label displayed
        """
        self.time_ticks_total = 5
        """
        gaps: [[float, float], ] - list of [min, max] of gaps
        """
        self.gaps = []
        """
        gap_bar: matplotlib.axes.Axes - axes where gaps are plotted
        """
        self.gap_bar = None
        """
        timestamp_bar_top: matplotlib.axes.Axes - axes to show times that
            places at the top of all channel plots
        timestamp_bar_bottom: matplotlib.axes.Axes - axes to show times that
            places at the bottom of all channel plots
        """
        self.timestamp_bar_top = None
        self.timestamp_bar_bottom = None
        """
        bottom: float - y position of the bottom edge of all plots in self.axes
            This is to identify the bottom of rulers
        """
        self.bottom = 0
        """
        c_mode: str - 'B'/'W' - color modes black or white
        """
        self.c_mode = 'B'
        """
        display_color: dict - that defined colors to be used.
        """
        self.display_color = {}
        self.set_colors('B')

    # ======================================================================= #
    #                                  EVENT
    # ======================================================================= #
    def resizeEvent(self, event):
        """
        OVERRIDE Qt method.
        When plottingWidget's viewport is resized along with parent window
            (even opening MainWindow triggers resizeEvent too), self.ratio_w,
            self.plotting_w, self.plotting_l will be recalculated so that the
            drawn channels fit inside the width of the new viewport.
        If self.plot_total == 0, when there is no channels drawn, height
            is set to the height of viewport to cover the whole viewport since
            viewport and widget may be in different colors.

        :param event: QResizeEvent - resize event
        """
        geo = self.maximumViewportSize()

        # set view size fit with the scroll's view port size
        self.main_widget.setFixedWidth(geo.width())
        self.ratio_w = geo.width() / self.width_base_px
        self.plotting_w = self.ratio_w * self.width_base
        self.plotting_l = self.ratio_w * self.plotting_l_base
        if self.plot_total == 0:
            self.main_widget.setFixedHeight(geo.height())

        return super(PlottingWidget, self).resizeEvent(event)

    def get_timestamp(self, event):
        """
        Get timestamp which is xdata value from the coordinate data of the
            event. Normally, xdata is an attribute of event. However, when
            mouse falls into spaces in between axes, event.xdata is None.

        :param event: button_press_event - mouse click event
        """
        x, y = event.x, event.y  # coordinate data
        inv = self.axes[0].transData.inverted()
        # convert to timestamp, bc event.xdata is None in the space bw axes
        xdata = inv.transform((x, y))[0]
        return xdata

    def zoom_bw_markers(self, xdata):
        """
        Zoom the section in between zoom_marker1 and zoom_marker2:
            + set flag zoom_marker1_shown to False
            + set lim to the new curr_min_x and new curr_max_x set by x of
                zoom_marker1 and zoom_marker2
            + hide zoom_marker1 and zoom_marker2

        :param xdata: float - time value in plot
        """
        if self.min_x == xdata:
            return
        self.zoom_marker1_shown = False
        [self.min_x, self.max_x] = sorted([self.min_x, xdata])
        self.set_lim()
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)

    def on_button_press_event(self, event):
        """
        When click mouse on the current plottingWidget, SOHView will loop
            through different plottingWidgets to do the same task for
            interaction:
            + shift+click: call on_shift_click() to do zooming
            + ctrl+click or cmd+click in mac: call on_ctrl_cmd_click() to show
                ruler

        :param event: button_press_event - mouse click event
        """
        modifiers = event.guiEvent.modifiers()
        try:
            # on_pick() take place after on_pick_on_artist where
            # tps_t was assigned in TimePowerSquareWidget
            xdata = self.tps_t
        except AttributeError:
            xdata = self.get_timestamp(event)
        for w in self.peer_plotting_widgets:
            if modifiers == QtCore.Qt.ShiftModifier:
                w.on_shift_click(xdata)
            elif modifiers in [QtCore.Qt.ControlModifier,
                               QtCore.Qt.MetaModifier]:
                w.on_ctrl_cmd_click(xdata)
            w.draw()

    def on_ctrl_cmd_click(self, xdata):
        """
        On ctrl/cmd + left click
            + hide zoom_marker1
            + connect ruler to follow mouse
            + show ruler

        :param xdata: float - time value of a channel plot
        """
        self.zoom_marker1.set_visible(False)
        self.zoom_marker1_shown = False
        try:
            self.fig.canvas.mpl_disconnect(self.follower)
        except AttributeError:
            pass
        self.set_ruler_visibled(self.ruler, xdata)

    def on_shift_click(self, xdata):
        """
        On shift + left click:
        if zoom_marker1 not shown yet:
            + hide ruler
            + connect zoom_marker2 to follow mouse
            + show zoom_marker1
        else:
            + show zoom_marker2
            + zoom data in between 2 zoomMarkers

        :param xdata: float - time value of a channel plot
        """
        if not self.zoom_marker1_shown:
            self.ruler.set_visible(False)
            self.set_ruler_visibled(self.zoom_marker1, xdata)
            self.min_x = xdata
            self.zoom_marker1_shown = True
            self.set_ruler_visibled(self.zoom_marker2, xdata)
        else:
            self.zoom_bw_markers(xdata)

    def set_ruler_visibled(self, ruler, xdata):
        """
        When set ruler visibled, x of positions of top and bottom have to
            follow the given x.
        If ruler is zoom_marker2, set it to follow the x of mouse's position.

        :param ruler: ConnectionPatch - one of ruler, zoom_marker1,
            zoom_marker2 defined in __init__
        :param xdata: float - time value of a channel plot
        """

        ruler.set_visible(True)
        ruler.xy1 = (xdata, 0)
        ruler.xy2 = (xdata, self.bottom)
        try:
            if ruler == self.zoom_marker2:
                # make zoom_marker2 follow mouse.
                # need to disconnect when state of rulers change
                self.follower = self.fig.canvas.mpl_connect(
                    "motion_notify_event", self.zoom_marker2_follow_mouse)
        except AttributeError:
            pass

    def zoom_marker2_follow_mouse(self, mouseevent):
        """
        Set zoom_marker2's to follow mouse's x.

        :param mouseevent: motion_notify_event - event to help keeping track
            of mouse move
        """
        xdata = self.get_timestamp(mouseevent)
        self.zoom_marker2.xy1 = (xdata, 0)
        self.zoom_marker2.xy2 = (xdata, self.bottom)
        self.draw()

    def keyPressEvent(self, event):
        """
        OVERRIDE Qt method.
        When press on Escape key, hide all rulers and set False for
            zoom_marker1_shown on all plotting widgets

        :param event: QKeyEvent - event to know what key is pressed
        """
        if event.key() == QtCore.Qt.Key_Escape:
            for w in self.peer_plotting_widgets:
                w.set_rulers_invisible()
                w.zoom_marker1_shown = False
                w.draw()
        return super(PlottingWidget, self).keyPressEvent(event)

    def set_rulers_invisible(self):
        """
        Hide ruler, zoom_marker1, zoom_marker2
        """
        self.ruler.set_visible(False)
        self.zoom_marker1.set_visible(False)
        self.zoom_marker2.set_visible(False)

    # ============================= END EVENT ==============================

    def add_timestamp_bar(self, height, top=True):
        """
        Set the axes to display timestamp_bar including color, ticks' size,
            label.

        :param height: float - height of timestamp_bar
        :param top: bool - flag indicating this timestamp_bar is located on top
            or bottom to locate tick label.
        """
        self.plotting_bot -= height
        timestamp_bar = self.canvas.figure.add_axes(
            [self.plotting_l, self.plotting_bot, self.plotting_w, 0.00005],
        )
        # Some plotting widgets like TPS doesn't have timestamp_bar showed
        # but still need it to locate the title of the plot => set it to off
        timestamp_bar.axis('off')
        timestamp_bar.xaxis.set_minor_locator(AutoMinorLocator())
        timestamp_bar.spines['bottom'].set_color(self.display_color['basic'])
        timestamp_bar.spines['top'].set_color(self.display_color['basic'])

        if top:
            labelbottom = False
        else:
            labelbottom = True
        timestamp_bar.tick_params(which='major', length=7, width=2,
                                  direction='inout',
                                  colors=self.display_color['basic'],
                                  labelbottom=labelbottom,
                                  labeltop=not labelbottom)
        timestamp_bar.tick_params(which='minor', length=4, width=1,
                                  direction='inout',
                                  colors=self.display_color['basic'])
        timestamp_bar.set_ylabel('Hours',
                                 fontweight='bold',
                                 fontsize=self.font_size,
                                 rotation=0,
                                 labelpad=constants.HOUR_TO_TMBAR_D,
                                 ha='left',
                                 color=self.display_color['basic'])
        # not show any y ticks
        timestamp_bar.set_yticks([])
        return timestamp_bar

    def update_timestamp_bar(self, timestamp_bar):
        """
        Update major, minor x ticks, tick labels on timestamp_bar based on
        curr_min_x, curr_max_x, date_mode, time_ticks_total, font_size.

        :param timestamp_bar: matplotlib.axes.Axes - axes for timestamp_bar
        """
        times, major_times, major_time_labels = getTimeTicks(
            self.min_x, self.max_x, self.date_mode,
            self.time_ticks_total
        )
        timestamp_bar.axis('on')
        timestamp_bar.set_xticks(times, minor=True)
        timestamp_bar.set_xticks(major_times)
        timestamp_bar.set_xticklabels(major_time_labels,
                                      fontsize=self.font_size + 2)
        timestamp_bar.set_xlim(self.min_x, self.max_x)

    def create_axes(self, plot_b, plot_h, has_min_max_lines=True):
        """
        Create axes to plot a channel.

        :param plot_b: float - bottom of the plot
        :param plot_h: float - height of the plot
        :param has_min_max_lines: bool - flag showing if the plot need min/max
            lines
        :return ax: matplotlib.axes.Axes - axes of a channel
        """
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, plot_b, self.plotting_w, plot_h],
            picker=True
        )

        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        if has_min_max_lines:
            ax.spines['top'].set_zorder(constants.Z_ORDER['AXIS_SPINES'])
            ax.spines['bottom'].set_zorder(constants.Z_ORDER['AXIS_SPINES'])
            ax.spines['top'].set_color(self.display_color['sub_basic'])
            ax.spines['bottom'].set_color(self.display_color['sub_basic'])

        # no stick (use timestamp bar as stick)
        ax.set_yticks([])
        ax.set_xticks([])
        # set tick_params for left ticklabel
        ax.tick_params(colors=self.display_color['basic'],
                       width=0,
                       pad=-2,
                       labelsize=self.font_size)
        # transparent background => self.fig will take care of background
        ax.patch.set_alpha(0)
        return ax

    def set_axes_info(self, ax, sample_no_list,
                      label=None, info='', y=None, chan_db_info=None,
                      linked_ax=None):
        """
        Draw plot's title, sub title, sample total label, center line, y labels
        for a channel.

        :param ax: matplotlib.axes.Axes - axes of a channel
        :param sample_no_list: [int,] - list of totals of different sample
            groups
        :param label: str/None - title of the plot.
            If None, show chan_db_info['label']
        :param info: str - additional info to show in sub title which is
            smaller and under title on the left side
        :param y: numpy.array - y values of the channel, to show min/max labels
            and min/max lines
        :param chan_db_info: dict - info of channel from database
        :param linked_ax: matplotlib.axes.Axes/None -
            if linked_ax is None, this is a main channel, label of channel will
                be displayed with title's format, on top right of plot.
            if linked_ax is not None, this is a channel using main channel's
                axes, label of channel will be displayed with sub title's
                format - under main title.
        """
        if label is None:
            label = chan_db_info['label']
        title_ver_alignment = 'center'
        # set info in subtitle under title
        if linked_ax is not None:
            info = label
        if info != '':
            ax.text(
                -0.15, 0.2,
                info,
                horizontalalignment='left',
                verticalalignment='top',
                rotation='horizontal',
                transform=ax.transAxes,
                color=self.display_color['sub_basic'],
                size=self.font_size
            )
            title_ver_alignment = 'top'

        if linked_ax is None:
            # set title on left side
            color = self.display_color['plot_label']
            if label.startswith("DEFAULT"):
                color = self.display_color["warning"]
            ax.text(
                -0.15, 0.6,
                label,
                horizontalalignment='left',
                verticalalignment=title_ver_alignment,
                rotation='horizontal',
                transform=ax.transAxes,
                color=color,
                size=self.font_size + 2
            )

        # set samples' total on right side
        if len(sample_no_list) == 1:
            ax.sampleLbl = ax.text(
                1.005, 0.5,
                sample_no_list[0],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                color=self.display_color['basic'],
                size=self.font_size
            )
        else:
            # bottom
            ax.sampleLbl = ax.text(
                1.005, 0.25,
                sample_no_list[0],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                color=self.display_color['basic'],
                size=self.font_size
            )
            # top
            ax.sampleLbl = ax.text(
                1.005, 0.75,
                sample_no_list[1],
                horizontalalignment='left',
                verticalalignment='center',
                rotation='horizontal',
                transform=ax.transAxes,
                color=self.display_color['basic'],
                size=self.font_size
            )

        if y is None:
            # draw center line
            ax.plot([self.min_x, self.max_x],
                    [0, 0],
                    color=self.display_color['sub_basic'],
                    linewidth=0.5,
                    zorder=constants.Z_ORDER['CENTER_LINE']
                    )
            ax.spines['top'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
        else:
            min_y = y.min()
            max_y = y.max()
            ax.spines['top'].set_visible(True)
            ax.spines['bottom'].set_visible(True)
            ax.unit_bw = getUnitBitweight(
                chan_db_info, self.parent.bit_weight_opt
            )
            self.set_axes_ylim(ax, min_y, max_y)

    def set_axes_ylim(self, ax, min_y, max_y):
        """
        Limit y range in min_y, max_y.
        Set y tick labels at min_y, max_y
        :param ax: matplotlib.axes.Axes - axes of a channel
        :param min_y: float - minimum of y values
        :param max_y: float - maximum of y values
        """
        min_y = round(min_y, 7)
        max_y = round(max_y, 7)
        if max_y > min_y:
            # There are different values for y => show yticks for min, max
            # separately
            ax.set_yticks([min_y, max_y])
            ax.set_yticklabels(
                [ax.unit_bw.format(min_y), ax.unit_bw.format(max_y)])
        if min_y == max_y:
            # All values for y are the same => show only one yticks
            max_y += 1
            ax.set_yticks([min_y])
            ax.set_yticklabels([ax.unit_bw.format(min_y)])
        ax.set_ylim(min_y, max_y)

    def add_gap_bar(self, gaps):
        """
        Draw the axes and labels to display gap_bar right under top timestamp
        bar.
        Update plotting_bot to draw next plot.

        :param gaps: [[float, float], ] - list of [min, max] of gaps
        """
        if self.parent.min_gap is None:
            return
        self.gaps = getGaps(gaps, float(self.parent.min_gap))
        self.plotting_bot -= 0.003
        self.gap_bar = self.create_axes(self.plotting_bot,
                                        0.001,
                                        has_min_max_lines=False)

        gap_label = "%sm" % self.parent.min_gap
        h = 0.001  # height of rectangle represent gap
        self.set_axes_info(self.gap_bar, [len(self.gaps)],
                           label=gap_label)
        # draw gaps
        for i in range(len(self.gaps)):
            x = self.gaps[i][0]
            # width of rectangle represent gap
            w = self.gaps[i][1] - self.gaps[i][0]
            self.gap_bar.add_patch(
                Rectangle(
                    (x, - h / 2), w, h, color='r', picker=True, lw=0.,
                    zorder=constants.Z_ORDER['GAP']
                )
            )

    def get_height(self, ratio, bw_plots_distance=0.0015):
        """
        Calculate new plot's bottom position and return plot's height.

        :param ratio: float - ratio of the plot height on the BASIC_HEIGHT
        :param bw_plots_distance: float - distance between plots
        :return plot_h: float - height of the plot
        """
        plot_h = constants.BASIC_HEIGHT * ratio  # ratio with figure height
        self.plotting_bot -= plot_h + bw_plots_distance
        self.plotting_bot_pixel += 19 * ratio
        return plot_h

    def plot_none(self):
        """
        Plot with nothing needed to show rulers.
        :return ax: matplotlib.axes.Axes - axes of the empty plot
        """
        plot_h = 0.00001
        bw_plots_distance = 0.0001
        self.plotting_bot -= plot_h + bw_plots_distance
        ax = self.create_axes(self.plotting_bot, plot_h,
                              has_min_max_lines=False)
        ax.x = None
        ax.plot([0], [0], linestyle="")
        return ax

    def plot_multi_color_dots(self, c_data, chan_db_info, chan_id,
                              ax, linked_ax):
        """
        plot scattered dots with colors defined by valueColors in database:
            Ex: *:W  means everything with white color
            Ex: -1:_|0:R|2.3:Y|+2.3:G  means
                value <= -1  => not plot
                value <= 0   => plot with R color
                value <= 2.3 => plot with Y color
                value > 2.3  => plot with G color
        Color codes are defined in colorSettings and limitted in 'valColRE'
            in dbSettings.py

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db_info['height'])
            ax = self.create_axes(
                self.plotting_bot, plot_h, has_min_max_lines=False)

        x = []
        prev_val = -constants.HIGHEST_INT

        if chan_db_info['valueColors'] in [None, 'None', '']:
            chan_db_info['valueColors'] = '*:W'
        value_colors = chan_db_info['valueColors'].split('|')
        for vc in value_colors:
            v, c = vc.split(':')
            val = getVal(v)
            if c == '_':
                prev_val = val
                continue

            if v.startswith('+'):
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if c_data['data'][i] > val]
            elif v == '*':
                points = c_data['times']
            else:
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if prev_val < c_data['data'][i] <= val]
            x += points

            ax.plot(points, len(points) * [0], linestyle="",
                    marker='s', markersize=0.5,
                    zorder=constants.Z_ORDER['DOT'],
                    color=Clr[c], picker=True, pickradius=3)
            prev_val = val

        total_samples = len(x)

        x = sorted(x)
        self.set_axes_info(ax, [total_samples], chan_db_info=chan_db_info,
                           linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_up_down_dots(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot channel with 2 different values, one above, one under center line.
        Each value has corresponding color defined in valueColors in database.
        Ex: 1:Y|0:R  means
            value == 1 => plot above center line with Y color
            value == 0 => plot under center line with R color
        Color codes are defined in colorSettings

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db_info['height'])
            ax = self.create_axes(
                self.plotting_bot, plot_h, has_min_max_lines=False)

        val_cols = chan_db_info['valueColors'].split('|')
        points_list = []
        colors = []
        for vc in val_cols:
            v, c = vc.split(':')
            val = getVal(v)

            points = [c_data['times'][i]
                      for i in range(len(c_data['data']))
                      if c_data['data'][i] == val]
            points_list.append(points)
            colors.append(c)

        # down dots
        ax.plot(points_list[0], len(points_list[0]) * [-0.5], linestyle="",
                marker='s', markersize=2, zorder=constants.Z_ORDER['DOT'],
                color=Clr[colors[0]], picker=True, pickradius=3)
        # up dots
        ax.plot(points_list[1], len(points_list[1]) * [0.5], linestyle="",
                marker='s', markersize=2, zorder=constants.Z_ORDER['DOT'],
                color=Clr[colors[1]], picker=True, pickradius=3)
        x = points_list[0] + points_list[1]
        x = sorted(x)
        ax.set_ylim(-2, 2)
        self.set_axes_info(ax, [len(points_list[0]), len(points_list[1])],
                           chan_db_info=chan_db_info,
                           linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_time_dots(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot times only
        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]

        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db_info['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        color = 'W'
        if chan_db_info['valueColors'] not in [None, 'None', '']:
            color = chan_db_info['valueColors'].strip()
        x = c_data['times']
        self.set_axes_info(ax, [len(x)], chan_db_info=chan_db_info,
                           linked_ax=linked_ax)

        ax.myPlot = ax.plot(x, [0] * len(x), marker='s', markersize=1.5,
                            linestyle='', zorder=constants.Z_ORDER['LINE'],
                            color=Clr[color], picker=True,
                            pickradius=3)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_lines_dots(self, c_data, chan_db_info, chan_id,
                        ax, linked_ax, info=''):
        """
        Plot lines with dots at the data points. Colors of dot and lines are
        defined in valueColors in database.
        Ex: L:G|D:W  means
            Lines are plotted with color G
            Dots are plotted with color W
        If D is not defined, dots won't be displayed.
        If L is not defined, lines will be plotted with color G
        Color codes are defined in colorSettings

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id] or DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id] for waveform data
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :param info: str - additional info to be displayed on sub-title under
            main-title
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.get_height(chan_db_info['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        x, y = c_data['times'], c_data['data']
        self.set_axes_info(ax, [len(x)], chan_db_info=chan_db_info,
                           info=info, y=y, linked_ax=linked_ax)
        colors = {}
        if chan_db_info['valueColors'] not in [None, 'None', '']:
            color_parts = chan_db_info['valueColors'].split('|')
            for cStr in color_parts:
                obj, c = cStr.split(':')
                colors[obj] = c

        l_color = 'G'
        has_dot = False
        if 'L' in colors:
            l_color = colors['L']
        if 'D' in colors:
            d_color = colors['D']
            has_dot = True

        if not has_dot:
            ax.myPlot = ax.plot(x, y,
                                linestyle='-', linewidth=0.7,
                                color=Clr[l_color])
        else:
            ax.myPlot = ax.plot(x, y, marker='s', markersize=1.5,
                                linestyle='-', linewidth=0.7,
                                zorder=constants.Z_ORDER['LINE'],
                                color=Clr[l_color],
                                markerfacecolor=Clr[d_color],
                                picker=True, pickradius=3)
        if linked_ax is None:
            ax.x = x
            ax.y = y
        else:
            ax.linkedX = x
            ax.linkedY = y
        return ax

    def plot_lines_s_rate(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot line only for waveform data channel (seismic data). Sample rate
        unit will be displayed

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if c_data['samplerate'] >= 1.0:
            info = "%dsps" % c_data['samplerate']
        else:
            info = "%gsps" % c_data['samplerate']
        return self.plot_lines_dots(c_data, chan_db_info, chan_id,
                                    ax, linked_ax, info=info)

    def plot_lines_mass_pos(self, c_data, chan_db_info, chan_id,
                            ax, linked_ax):
        """
        Plot multi-color dots with grey line for mass position channel.
        Use getMassposValueColors() to get value_colors map based on
            Menu - MP Coloring selected from Main Window.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            mass_pos_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        value_colors = getMassposValueColors(
            self.parent.mass_pos_volt_range_opt, chan_id,
            self.c_mode, self.processing_log, retType='tupleList')

        if value_colors is None:
            return

        if ax is None:
            plot_h = self.get_height(chan_db_info['height'])
            ax = self.create_axes(self.plotting_bot, plot_h)

        ax.x, ax.y = c_data['times'], c_data['data']
        self.set_axes_info(ax, [len(ax.x)], chan_db_info=chan_db_info, y=ax.y)
        ax.myPlot = ax.plot(ax.x, ax.y,
                            linestyle='-', linewidth=0.7,
                            color=self.display_color['sub_basic'],
                            zorder=constants.Z_ORDER['LINE'])[0]
        colors = [None] * len(ax.y)
        sizes = [0.5] * len(ax.y)
        for i in range(len(ax.y)):
            count = 0
            prev_v = 0
            for v, c in value_colors:
                if count < (len(value_colors) - 1):
                    if prev_v < abs(ax.y[i]) <= v:
                        colors[i] = Clr[c]
                        break
                else:
                    colors[i] = Clr[c]
                    break
                prev_v = v
                count += 1
        ax.scatter(ax.x, ax.y, marker='s', c=colors, s=sizes,
                   zorder=constants.Z_ORDER['DOT'])
        return ax

    def add_ruler(self, color):
        """
        Create a line across all channels' plots from top timestamp
            bar to bottom timestamp bar. There are 3 different rulers defined
            in __init__(): self.ruler, self.zoom_marker1, self.zoom_marker2.

        :param color: color of the ruler
        """
        ruler = ConnectionPatch(
            xyA=(0, 0),
            xyB=(0, self.bottom),
            coordsA="data",
            coordsB="data",
            axesA=self.timestamp_bar_top,
            axesB=self.timestamp_bar_bottom,
            color=color,
        )
        ruler.set_visible(False)
        self.timestamp_bar_bottom.add_artist(ruler)
        return ruler

    def set_lim(self, first_time=False):
        """
        + Re-decimate channel data to get more detail when zoom in
            (get_zoom_data)
        + Update timestamp bar with new ticks info
        + Update gap_bar by setting xlim and re-calculate gap total.
        + for each axes, set new x_lim, y_lim, label for totals of data points

        :param first_time: bool - flag shows that this set_lim is called the
            fist time for this data set or not.
        """
        if not first_time:
            for chanID in self.plotting_data1:
                c_data = self.plotting_data1[chanID]
                self.get_zoom_data(c_data, chanID)
            for chanID in self.plotting_data2:
                c_data = self.plotting_data2[chanID]
                self.get_zoom_data(c_data, chanID)
        self.update_timestamp_bar(self.timestamp_bar_top)
        self.update_timestamp_bar(self.timestamp_bar_bottom)
        if self.gap_bar is not None:
            self.gap_bar.set_xlim(self.min_x, self.max_x)
            if not first_time:
                new_gaps = [g for g in self.gaps
                            if (self.min_x <= g[0] <= self.max_x
                                or self.min_x <= g[1] <= self.max_x)]

                # reset total of samples on the right
                self.gap_bar.sampleLbl.set_text(len(new_gaps))
        for ax in self.axes:
            ax.set_xlim(self.min_x, self.max_x)
            if ax.x is None:
                # the plot_none bar is at the end, no need to process
                break
            if not first_time:
                # x, y
                new_x_indexes = np.where((ax.x >= self.min_x) &
                                         (ax.x <= self.max_x))

                # reset total of samples on the right
                ax.sampleLbl.set_text(len(new_x_indexes))

                if len(new_x_indexes) == 0:
                    continue
                if hasattr(ax, 'y'):
                    # don't need to reset y range if ax.y not exist
                    new_x = ax.x[new_x_indexes]
                    new_min_x = min(new_x)
                    new_max_x = max(new_x)
                    try:
                        new_min_x_index = ax.x.index(new_min_x)
                        new_max_x_index = ax.x.index(new_max_x)
                    except AttributeError:
                        new_min_x_index = np.where(ax.x == new_min_x)[0][0]
                        new_max_x_index = np.where(ax.x == new_max_x)[0][0]
                    new_y = ax.y[new_min_x_index:new_max_x_index + 1]
                    new_min_y = min(new_y)
                    new_max_y = max(new_y)
                    self.set_axes_ylim(ax, new_min_y, new_max_y)

    def set_title(self, title, x=-0.15, y=100, v_align='top'):
        """
        Display title of the data set's plotting based on

        :param title: str - title of the data set's plotting
        :param x: float - horizontal position to the right from the left edge
            of the self.timestamp_bar_top
        :param y: float - vertical position upward from the top edge of
            the self.timestamp_bar_top
        :param v_align: str - vertical alignment of title to the
            self.timestamp_bar_top
        """
        self.fig.text(x, y, title,
                      verticalalignment=v_align,
                      horizontalalignment='left',
                      transform=self.timestamp_bar_top.transAxes,
                      color=self.display_color['basic'],
                      size=self.font_size)

    def draw(self):
        """
        Update drawing on the widget.
        """
        try:
            self.canvas.draw()
            # a bug on mac:
            # not showing updated info until clicking on another window
            # fix by calling repaint()
            self.main_widget.repaint()
        except TypeError:
            pass

    # ######## Functions for outside world #####
    def init_size(self):
        """
        Set widget to fit with the view port when window first open.
        """
        geo = self.maximumViewportSize()
        if self.plot_total == 0:
            # set view size fit with the scroll's view port size
            self.main_widget.setFixedWidth(geo.width())
            self.main_widget.setFixedHeight(geo.height())

    # =============================== axes ================================
    def plot_channels(self, start_tm, end_tm, key, data_time,
                      channel_list, time_ticks_total,
                      soh_data, mass_pos_data, gaps):
        """
        Recursively plot each channels for soh_data and masspos_data.

        :param start_tm: float - requested start time to read
        :param end_tm: float - requested end time to read
        :param key: str or (str, str) station name  for mseed,
            or (device's experiment number, experiment number ) for reftek
        :param data_time: [float, float] - time range of the selected data set
        :param channel_list: [str,] - SOH channel list read from the selected
            data set
        :param time_ticks_total: int - max number of tick to show on time bar
        :param soh_data: dict - SOH data of the selected data set, refer to
            DataTypeModel.__init__.soh_data[key]
        :param mass_pos_data: dict - mass position data of the selected
            data set, refer to DataTypeModel.__init__.mass_pos_data[key]
         :param gaps: [[float, float],] - list of gaps of the selected data set
        """
        self.plotting_data1 = soh_data
        self.plotting_data2 = mass_pos_data
        self.processing_log = []
        self.gap_bar = None

        if self.axes:
            self.fig.clear()
        self.date_mode = self.parent.date_format.upper()
        self.time_ticks_total = time_ticks_total

        self.min_x = max(data_time[0], start_tm)
        self.max_x = min(data_time[1], end_tm)
        self.plot_total = len(self.plotting_data1) + len(self.plotting_data2)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = constants.BOTTOM
        self.plotting_bot_pixel = constants.BOTTOM_PX
        self.axes = []

        self.timestamp_bar_top = self.add_timestamp_bar(0.003)
        self.set_title(title)
        self.add_gap_bar(gaps)
        not_found_chan = [c for c in channel_list
                          if c not in self.plotting_data1.keys()]
        if len(not_found_chan) > 0:
            msg = (f"The following channels is in Channel Preferences but "
                   f"not in the given data: {not_found_chan}")
            self.processing_log.append((msg, 'warning'))

        for chan_id in self.plotting_data1:
            chan_db_info = extractData.getChanPlotInfo(chan_id,
                                                       self.parent.data_type)
            if chan_db_info['height'] == 0:
                # not draw
                continue
            if chan_db_info['channel'] == 'DEFAULT':
                msg = (f"Channel {chan_id}'s "
                       f"definition can't be found database.")
                displayTrackingInfo(self.tracking_box, msg, 'warning')

            if chan_db_info['plotType'] == '':
                continue
            self.plotting_data1[chan_id]['chan_db_info'] = chan_db_info
            self.get_zoom_data(self.plotting_data1[chan_id], chan_id,
                               self.plotting_data1, True)

        for chan_id in self.plotting_data2:
            chan_db_info = extractData.getChanPlotInfo(
                chan_id, self.parent.data_type)
            self.plotting_data2[chan_id]['chan_db_info'] = chan_db_info
            self.get_zoom_data(self.plotting_data2[chan_id], chan_id, True)

        self.axes.append(self.plot_none())
        self.timestamp_bar_bottom = self.add_timestamp_bar(0.003, top=False)
        self.set_lim(first_time=True)
        self.bottom = self.axes[-1].get_ybound()[0]
        self.ruler = self.add_ruler(self.display_color['time_ruler'])
        self.zoom_marker1 = self.add_ruler(self.display_color['zoom_marker'])
        self.zoom_marker2 = self.add_ruler(self.display_color['zoom_marker'])
        # Set view size fit with the given data
        if self.main_widget.geometry().height() < self.plotting_bot_pixel:
            self.main_widget.setFixedHeight(self.plotting_bot_pixel)

        self.draw()

    def get_zoom_data(self, c_data, chan_id, plotting_data=None,
                      first_time=False):
        """
        Trim data of a channel to the zooming time range and re-downsample
            the data to get more detail.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to
            DataTypeModel.__init__.soh_data[key][chan_id] or
            DataTypeModel.__init__.mass_pos_data[key][chan_id]
        :param chan_id: str - name of channel
        :param plotting_data: dict - DataTypeModel.__init__.soh_data[key]. To
            help combine 2 channels in one axes in need
        :param first_time: bool - flag identify when the data set is first
            plotted, not zoom in yet
        """
        chan_db_info = c_data['chan_db_info']
        plot_type = chan_db_info['plotType']
        trim_downsample_SOHChan(c_data, self.min_x, self.max_x,
                                first_time)
        apply_convert_factor(c_data, 1)
        if 'ax' not in c_data:
            linked_ax = None
            if chan_db_info['linkedChan'] not in [None, 'None', '']:
                linked_ax = plotting_data[chan_db_info['linkedChan']]['ax']

            ax = getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db_info, chan_id, None, linked_ax)
            if ax is None:
                return
            c_data['ax'] = ax
            ax.chan = chan_id
            self.axes.append(ax)
        else:
            getattr(self, dbConf['plotFunc'][plot_type][1])(
                c_data, chan_db_info, chan_id, c_data['ax'], None)

    # ============================= END AXES ==============================
    def set_colors(self, mode):
        """
        Set color mode, colors' definition, background color for the graphic.

        param mode: str - 'B'/'W' - color modes black or white
        """
        self.c_mode = mode
        self.display_color = set_color_mode(mode)
        self.fig.patch.set_facecolor(self.display_color['background'])

    def set_peer_plotting_widgets(self, widgets):
        """
        Set other plotting widgets so that their rulers and zoom task can work
            synchronously.
        :param widgets: [PlottingWidgets,] - List of other plotting widgets
        """
        self.peer_plotting_widgets = widgets
